import numpy as np
from .problem import AbstractKOProblem, AbstractOUProblem
from ..simulation import SStatus


class RKOProblem(AbstractKOProblem):
    """
    Reaction Knockout Optimization Problem.

    :param model: The constraint metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        super(RKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)

    def _build_target_list(self):
        """Default modification target builder.
        Removes drains, transport and essential reactions
        """
        print("Building modification target list.")
        reactions = set(self.simulator.reactions)
        essential = set(self.simulator.essential_reactions())
        drains = set(self.simulator.get_exchange_reactions())
        transport = set(self.simulator.get_transport_reactions())
        target = reactions - essential - drains - transport
        if self.non_target is not None:
            target = target - set(self.non_target)
        target = list(target)
        self._trg_list = target


class ROUProblem(AbstractOUProblem):
    """
    Reaction Over/Under Expression Optimization Problem

    :param model: The constraint metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
    :param list levels: Over/under expression levels (Default EAConstants.LEVELS)
    :param boolean twostep: If deletions should be applied before identifiying reference flux values.
    """

    def __init__(self, model, fevaluation=None, **kwargs):
        super(ROUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)

    def _build_target_list(self):
        print("Building modification target list.")
        reactions = set(self.simulator.reactions)
        # drains = set(self.simulator.get_exchange_reactions())
        target = reactions  # - drains
        if self.non_target is not None:
            target = target - set(self.non_target)
        target = list(target)
        self._trg_list = target

    def solution_to_constraints(self, candidate):
        """
        Decodes a candidate, an dict {idx:lv} into a dictionary of constraints
        Suposes that reverseble reactions have been treated and bounded with positive flux values
        """
        constraints = dict()
        # computes reference fluxes based on deletions
        reference = self.reference
        if self.twostep:
            try:
                deletions = {rxn: 0 for rxn, lv in candidate.items() if lv == 0}
                sr = self.simulator.simulate(constraints=deletions, method='pFBA')
                if sr.status in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
                    reference = sr.fluxes
            except Exception as e:
                print(e)

        for rxn, lv in candidate.items():
            rev_rxn = self.simulator.reverse_reaction(rxn)
            # skips if the reverse reaction was already processed
            if rev_rxn and rev_rxn in constraints.keys():
                continue
            elif lv < 0:
                raise ValueError("All UO levels should be positive")
            else:
                constraints.update(self.reaction_constraints(rxn, lv, reference))
        return constraints


class MediumProblem(AbstractOUProblem):
    """
    Medium Optimization Problem. Try to find an optimized uptake configuration.
    By default all uptake reactions are considered. Uptake reactions not included in
    a solution candidate are KO.

    :param model: The constraint metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
    :param list levels: Over/under expression levels (Default [0,0.1,0.2,...,9.9,10.0])

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        super(MediumProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        self.levels = kwargs.get('levels', np.linspace(0, 10, 101))
        self.candidate_max_size = kwargs.get(
            'candidate_max_size', len(self.target_list))
        self.simulator._allow_env_changes = True

    def _build_target_list(self):
        target = set(self.simulator.get_uptake_reactions())
        if self.non_target is not None:
            target = target - set(self.non_target)
        target = list(target)
        self._trg_list = target

    def solution_to_constraints(self, candidate):
        """
        Decodes a candidate, a dict {idx:lv} into a dictionary of constraints
        Suposes that reversible reactions have been treated and bounded with positive flux values
        """
        constraints = dict()
        from mewpy.util.constants import ModelConstants
        for rxn in self.target_list:
            if rxn in candidate.keys():
                lv = candidate[rxn]
                constraints[rxn] = (-1 * lv, ModelConstants.REACTION_UPPER_BOUND)
            else:
                constraints[rxn] = (0, 0)
        return constraints



""" Adapted from Daniel Machado's REFRAMED"""
from mewpy.simulation import SStatus
from mewpy.simulation.simulation import Simulator, SimulationResult
from enum import Enum
import re


class Status(Enum):
    """ Enumeration of possible solution status. """
    OPTIMAL = 'Optimal'
    UNKNOWN = 'Unknown'
    SUBOPTIMAL = 'Suboptimal'
    UNBOUNDED = 'Unbounded'
    INFEASIBLE = 'Infeasible'
    INF_OR_UNB = 'Infeasible or Unbounded'

status_mapping={
    Status.OPTIMAL : SStatus.OPTIMAL,
    Status.UNKNOWN : SStatus.UNKNOWN,
    Status.SUBOPTIMAL : SStatus.SUBOPTIMAL,
    Status.UNBOUNDED : SStatus.UNBOUNDED,
    Status.INFEASIBLE : SStatus.INFEASIBLE,
    Status.INF_OR_UNB : SStatus.INF_OR_UNB
}

class Solution(object):
    """ Stores the results of an optimization.

    Instantiate without arguments to create an empty Solution representing a failed optimization.
    """

    def __init__(self, status=Status.UNKNOWN, message=None, fobj=None, values=None,
                 shadow_prices=None, reduced_costs=None):
        self.status = status
        self.message = message
        self.fobj = fobj
        self.values = values
        self.shadow_prices = shadow_prices
        self.reduced_costs = reduced_costs

    def __str__(self):
        return f"Objective: {self.fobj}\nStatus: {self.status.value}\n"

    def __repr__(self):
        return str(self)

    def to_dataframe(self):
        """ Convert reaction fluxes to *pandas.DataFrame*

        Returns:
            pandas.DataFrame: flux values
        """
        try:
            import pandas as pd
        except ImportError:
            raise RuntimeError("Pandas is not installed.")

        return pd.DataFrame(self.values.values(), columns=["value"], index=self.values.keys())

def to_simulation_result(model, objective_value, constraints, sim, solution, method=None):
    res = SimulationResult(model.model if isinstance(model, Simulator) else model,
                           objective_value,
                           status= status_mapping[solution.status],
                           fluxes=solution.values,
                           envcond=sim.environmental_conditions,
                           model_constraints=sim._constraints.copy(),
                           simul_constraints=constraints,
                           method=method
                           )                           
    return res

def print_values(value_dict, pattern=None, sort=False, abstol=1e-9):

    values = [(key, value) for key, value in value_dict.items() if abs(value) > abstol]

    if pattern:
        re_expr = re.compile(pattern)
        values = [x for x in values if re_expr.search(x[0]) is not None]

    if sort:
        values.sort(key=lambda x: x[1])

    entries = (f'{r_id:<12} {val: .6g}'for (r_id, val) in values)

    print('\n'.join(entries))



import copy
from abc import ABC, abstractmethod

from .constants import EAConstants, ModelConstants


MP_Evaluators = []

from multiprocessing import Process
from multiprocessing.pool import Pool as MPPool
MP_Evaluators.append('nodaemon')

# pathos
try:
    import pathos.multiprocessing as multiprocessing
    from pathos.multiprocessing import Pool
    MP_Evaluators.append('mp')

except ImportError as e:
    import multiprocessing
    from multiprocessing.pool import Pool
    MP_Evaluators.append('mp')
# dask
try:
    import dask
    MP_Evaluators.append('dask')
except ImportError:
    pass
# pyspark
try:
    from pyspark import SparkConf, SparkContext
    MP_Evaluators.append('spark')
except ImportError:
    pass


class NoDaemonProcess(Process):

    def _get_daemon(self):
        return False
    def _set_daemon(self,value):
        pass
    daemon = property(_get_daemon, _set_daemon)

class NoDaemonProcessPool(MPPool):
    Process = NoDaemonProcess


def cpu_count():
    """ The number of cpus
        Return EAConstants.NUM_CPUS if it is set to a positive number
        otherwise return half of the available cpus.
    """
    if EAConstants.NUM_CPUS > 0:
        return EAConstants.NUM_CPUS
    else:
        try:
            return multiprocessing.cpu_count() // 2
        except (ImportError, NotImplementedError):
            return 1


class Evaluable(ABC):

    @abstractmethod
    def evaluator(self, candidates, *args):
        raise NotImplementedError


class Evaluator(ABC):

    """An interface for multiprocessing evaluators

    Raises:
        NotImplementedError: Requires an evaluated method to
        be implemented.
    """
    @abstractmethod
    def evaluate(self, candidates, args):
        raise NotImplementedError


class MultiProcessorEvaluator(Evaluator):

    def __init__(self, evaluator, mp_num_cpus):
        """A multiprocessing evaluator

        Args:
            evaluator(function): Evaluation function.
            mp_num_cpus(int): Number of CPUs
        """
        self.pool = Pool(mp_num_cpus)
        self.evaluator = evaluator
        self.__name__ = self.__class__.__name__

    def evaluate(self, candidates, args):
        """
        Values in args will be ignored and not passed to the evaluator to avoid unnecessary pickling in inspyred.
        """
        results = self.pool.map(self.evaluator, candidates)
        return results

    def __call__(self, candidates, args):
        return self.evaluate(candidates, args)


class NoDaemonMultiProcessorEvaluator(Evaluator):

    def __init__(self, evaluator, mp_num_cpus):
        """A multiprocessing evaluator

        Args:
            evaluator(function): Evaluation function.
            mp_num_cpus(int): Number of CPUs
        """
        self.pool = NoDaemonProcessPool(mp_num_cpus)
        self.evaluator = evaluator
        self.__name__ = self.__class__.__name__
        print('nodaemon')

    def evaluate(self, candidates, args):
        """
        Values in args will be ignored and not passed to the evaluator to avoid unnecessary pickling in inspyred.
        """
        results = self.pool.map(self.evaluator, candidates)
        return results

    def __call__(self, candidates, args):
        return self.evaluate(candidates, args)


class DaskEvaluator(Evaluator):

    def __init__(self, evaluator, mp_num_cpus, scheduler='processes'):
        """A Dask multiprocessing evaluator

        Args:
            evaluator (function): Evaluation function.
            mp_num_cpus (int): Number of CPUs.
        """
        self.evaluator = evaluator
        self.scheduler = scheduler
        self.__name__ = self.__class__.__name__

    def evaluate(self, candidates, args):
        with dask.config.set(scheduler=self.scheduler):
            return list(dask.compute(*[dask.delayed(self.evaluator)(c) for c in candidates]))

    def __call__(self, candidates, args):
        return self.evaluate(candidates, args)


class SparkEvaluator(Evaluator):
    def __init__(self, evaluator, mp_num_cpus):
        """A Spark multiprocessing evaluator

        Args:
            evaluator (function): Evaluation function.
            mp_num_cpus (int): Number of CPUs.
        """
        self.evaluator = evaluator
        self.spark_conf = SparkConf().setAppName(
            "mewpy").setMaster(f"local[{mp_num_cpus}]")
        self.spark_context = SparkContext(conf=self.spark_conf)
        self.__name__ = self.__class__.__name__

    def evaluate(self, candidates, args):
        solutions_to_evaluate = self.spark_context.parallelize(candidates)
        return solutions_to_evaluate.map(lambda s: self.evaluator(s)).collect()

    def __call__(self, candidates, args):
        return self.evaluate(candidates, args)


# ray
try:
    import ray
    MP_Evaluators.append('ray')
except ImportError:
    pass
else:

    @ray.remote
    class RayActor:
        """
        Each actor (worker) has a solver instance to overcome the need to serialize solvers which may not be pickable.
        The solver is not reset before each evaluation.
        """

        def __init__(self, problem):
            self.problem = copy.deepcopy(problem)

        def evaluate_candidates(self, candidates):
            """Evaluates a sublist of candidates
            """
            if getattr(self.problem, "evaluator", False):
                return self.problem.evaluator(candidates, None)
            else:
                res = []
                for candidate in candidates:
                    res.append(self.problem.evaluate(candidate))
                return res

    @ray.remote
    class RayActorF:
        """
        Each actor (worker) has a solver instance to overcome the need to serialize solvers which may not be pickable.
        The solver is not reset before each evaluation.
        """

        def __init__(self, func):
            self.func = copy.deepcopy(func)

        def evaluate_candidates(self, candidates):
            """Evaluates a sublist of candidates
            """
            res = []
            for candidate in candidates:
                res.append(self.func(candidate))
            return res

    class RayEvaluator(Evaluator):
        def __init__(self, problem, number_of_actors, isfunc=False):
            """A ray actor responsible for performing evaluations.

            Args:
                problem: A class implementing an evaluator(list_of_candidates,**kwargs)
                number_of_actors (int): Number of workers
            """
            ray.init(ignore_reinit_error=True)
            if isfunc:
                self.actors = [RayActorF.remote(problem)
                               for _ in range(number_of_actors)]
            else:
                self.actors = [RayActor.remote(problem)
                               for _ in range(number_of_actors)]
            self.number_of_actors = len(self.actors)
            self.__name__ = self.__class__.__name__
            print(f"Using {self.number_of_actors} workers.")

        def evaluate(self, candidates, args):
            """
            Divides the candidates into sublists to be evaluated by each actor
            """
            size = len(candidates) // self.number_of_actors
            n = len(candidates) % self.number_of_actors
            sub_lists = []
            p = 0
            for _ in range(self.number_of_actors):
                d = size + 1 if n > 0 else size
                sub_lists.append(candidates[p:p + d])
                p = p + d
                n -= 1
            values = []
            for i in range(self.number_of_actors):
                actor = self.actors[i]
                values.append(actor.evaluate_candidates.remote(sub_lists[i]))
            r = ray.get(values)
            result = []
            for x in r:
                result += x
            return result

        def __call__(self, candidates, args):
            return self.evaluate(candidates, args)


def get_mp_evaluators():
    """"Returns the list of available multiprocessing evaluators.
    """
    return MP_Evaluators


def get_evaluator(problem, n_mp=cpu_count(), evaluator=ModelConstants.MP_EVALUATOR):
    """Retuns a multiprocessing evaluator

    Args:
        problem: a class implementing an evaluate(candidate) function
        n_mp (int, optional): The number of cpus. Defaults to cpu_count().
        evaluator (str, optional): The evaluator name: options 'ray','dask','spark'.\
            Defaults to ModelConstants.MP_EVALUATOR.

    Returns:
        [type]: [description]
    """
    if evaluator == 'ray' and 'ray' in MP_Evaluators:
        return RayEvaluator(problem, n_mp)
    elif evaluator == 'nodaemon' and 'nodaemon' in MP_Evaluators:
        return NoDaemonMultiProcessorEvaluator(problem,n_mp) 
    elif evaluator == 'dask' and 'dask' in MP_Evaluators:
        return DaskEvaluator(problem.evaluate, n_mp)
    elif evaluator == 'spark' and 'spark' in MP_Evaluators:
        return SparkEvaluator(problem.evaluate, n_mp)
    else:
        return MultiProcessorEvaluator(problem.evaluate, n_mp)


def get_fevaluator(func, n_mp=cpu_count(), evaluator=ModelConstants.MP_EVALUATOR):
    """Retuns a multiprocessing evaluator

    Args:
        problem: a class implementing an evaluate(candidate) function
        n_mp (int, optional): The number of cpus. Defaults to cpu_count().
        evaluator (str, optional): The evaluator name: options 'ray','dask','spark'.\
            Defaults to ModelConstants.MP_EVALUATOR.

    Returns:
        [type]: [description]
    """
    if evaluator == 'ray' and 'ray' in MP_Evaluators:
        return RayEvaluator(func, n_mp, isfunc=True)
    elif evaluator == 'nodaemon' and 'nodaemon' in MP_Evaluators:
        return NoDaemonMultiProcessorEvaluator(func,n_mp) 
    elif evaluator == 'dask' and 'dask' in MP_Evaluators:
        return DaskEvaluator(func, n_mp)
    elif evaluator == 'spark' and 'spark' in MP_Evaluators:
        return SparkEvaluator(func, n_mp)
    else:
        return MultiProcessorEvaluator(func, n_mp)



from .ode import (ODEMethod, SolverConfigurations, ODEStatus,
                  KineticConfigurations)
from .solution import Solution, Status
from .sglobal import __MEWPY_solvers__, __MEWPY_ode_solvers__
# #################################################
# Linear Programming Solvers
# #################################################


default_solver = None


def get_default_solver():

    global default_solver

    if default_solver:
        return default_solver

    solver_order = ['cplex', 'gurobi', 'optlang']

    for solver in solver_order:
        if solver in list(__MEWPY_solvers__.keys()):
            default_solver = solver
            break

    if not default_solver:
        raise RuntimeError("No solver available.")

    return default_solver


def set_default_solver(solvername):
    """ Sets default solver.

    Arguments:
        solvername : (str) solver name (currently available: 'gurobi', 'cplex')
    """

    global default_solver

    if solvername.lower() in list(__MEWPY_solvers__.keys()):
        default_solver = solvername.lower()
    else:
        raise RuntimeError(f"Solver {solvername} not available.")


def solver_instance(model=None):
    """ Returns a new instance of the currently selected solver.

    Arguments:
        model : COBRApy/REFRAMED model or a Simulator (optional) -- immediatly instantiate problem with given model

    Returns:
        Solver
    """

    solver = get_default_solver()

    if solver:
        return __MEWPY_solvers__[solver](model)

# #################################################
# ODE solvers
# #################################################



try:
    from .scikits_solver import ScikitsODESolver
    __MEWPY_ode_solvers__['scikits'] = ScikitsODESolver
except ImportError:
    pass


try:
    from .scipy_solver import ScipySolver
    __MEWPY_ode_solvers__['scipy'] = ScipySolver
except ImportError:
    pass


try:
    from .odespy_solver import ODESpySolver
    __MEWPY_ode_solvers__['odespy'] = ODESpySolver
except ImportError:
    pass


default_ode_solver = None


def get_default_ode_solver():
    global default_ode_solver

    if default_ode_solver:
        return default_ode_solver

    ode_solver_order = ['scikits', 'scipy', 'odespy']

    for solver in ode_solver_order:
        if solver in list(__MEWPY_ode_solvers__.keys()):
            default_ode_solver = solver
            break

    if not default_ode_solver:
        raise RuntimeError("No solver ODE available.")

    return default_ode_solver


def set_default_ode_solver(solvername):
    """ Sets default solver.

    Arguments:
        solvername : (str) solver name (currently available: 'gurobi', 'cplex')
    """

    global default_ode_solver

    if solvername.lower() in list(__MEWPY_ode_solvers__.keys()):
        default_ode_solver = solvername.lower()
    else:
        raise RuntimeError(f"ODE solver {solvername} not available.")


def ode_solver_instance(func, method: ODEMethod):
    """ Returns a new instance of the currently selected solver.

    Arguments:
        func : a function
        method: a method

    Returns:
        Solver
    """

    solver = get_default_ode_solver()
    if solver:
        return __MEWPY_ode_solvers__[solver](func, method)



from typing import Union, TYPE_CHECKING, Type, Dict
import sys

from mewpy.germ.algebra import Expression
from mewpy.germ.algebra import parse_expression

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
    from mewpy.germ.variables import Variable, Gene, Interaction, Metabolite, Reaction, Regulator, Target

sys.setrecursionlimit(50000)


def serialize(serialization_name, deserialization_name=None, pickle_name=None):
    def wrapper(attr):
        attr.fget.serialize = serialization_name
        attr.fget.deserialize = deserialization_name
        attr.fget.pickle = pickle_name

        return attr

    return wrapper


class Serializer:

    # -----------------------------------------------------------------------------
    # Serializing to dictionary, tuple, string and int or float
    # -----------------------------------------------------------------------------

    @staticmethod
    def _regulatory_events_serializer(regulatory_events):

        return {key: expression.to_string() for key, expression in regulatory_events.items()}

    @staticmethod
    def _expression_serializer(expression):

        return expression.to_string()

    @staticmethod
    def _obj_serializer(obj):

        if isinstance(obj, set):

            return tuple(obj)

        elif hasattr(obj, 'id'):

            return obj.id

        return obj

    @staticmethod
    def _variable_container_serializer(container):

        return tuple(container.keys())

    @staticmethod
    def _key_container_serializer(container):

        return {key.id: value for key, value in container.items()}

    @staticmethod
    def _model_container_serializer(container):

        return {variable.id: variable.to_dict(serialization_format='json') for variable in container.values()}

    def _get_attribute_serializer(self, attr):

        if attr in ('id', 'name', 'types', 'aliases', 'target', 'charge', 'compartment', 'formula', 'compartments',
                    'bounds', 'coefficients', 'interaction'):

            return self._obj_serializer

        elif attr in ('reactions', 'regulators', 'genes', 'interactions', 'targets'):

            return self._variable_container_serializer

        elif attr == 'regulatory_events':

            return self._regulatory_events_serializer

        elif attr == 'gpr':

            return self._expression_serializer

        elif attr == 'stoichiometry':

            return self._key_container_serializer

        else:
            return lambda *args, **kwargs: {}

    def _variable_serializer(self: Union['Serializer', 'Variable', 'Model']):

        variable = {}

        for _, (attr, _, _) in self.attributes.items():
            serializer = self._get_attribute_serializer(attr=attr)

            attribute = getattr(self, attr)

            variable[attr] = serializer(attribute)

        return variable

    def _get_container_serializer(self, attr, variables=True):

        if attr in ('id', 'name', 'types'):

            return self._obj_serializer

        elif attr == 'objective':

            return self._key_container_serializer

        elif attr in ('genes', 'metabolites', 'reactions', 'interactions', 'regulators', 'targets'):

            if variables:
                return self._model_container_serializer

            return dict

        else:
            return lambda *args, **kwargs: {}

    def _model_serializer(self: Union['Serializer', 'Variable', 'Model'],
                          variables=True):

        model = {}

        for _, (attr, _, _) in self.containers.items():
            model[attr] = {}

            serializer = self._get_container_serializer(attr=attr, variables=variables)

            container = getattr(self, attr, {})

            model[attr] = serializer(container)

        return model

    # -----------------------------------------------------------------------------
    # Deserializing from dictionary, tuple, string and int or float
    # -----------------------------------------------------------------------------

    @classmethod
    def _model_deserializer(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                            obj,
                            variables=False):

        identifier = obj.get('id')
        types = obj.get('types')

        model = cls.from_types(types=types, identifier=identifier)

        if variables:
            children = cls._build_children(obj=obj, model=model)

        else:
            children = cls._get_children(obj=obj, model=model)

        update_attributes = {}
        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None:
                deserializer = cls._get_container_deserializer(attr=deserialize_name)

                container = obj[serialize_name]
                container = deserializer(container, children=children)

                if serialize_name in ('genes', 'metabolites', 'reactions', 'interactions', 'regulators', 'targets'):
                    setattr(model, f'_{serialize_name}', container)

                else:
                    update_attributes[deserialize_name] = container

        model.update(**update_attributes)

        return model

    @staticmethod
    def _build_children(obj, model):

        from mewpy.germ.variables import Variable

        children = {}

        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None and serialize_name in ('genes', 'metabolites', 'reactions',
                                                                   'interactions', 'regulators', 'targets'):

                container = obj[serialize_name]

                for var_id, variable in container.items():

                    if var_id in children:

                        continue

                    else:

                        children[var_id] = Variable.from_types(variable['types'],
                                                               identifier=variable['id'],
                                                               model=model)

        return children

    @staticmethod
    def _get_children(obj, model):

        children = {}

        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None and serialize_name in ('genes', 'metabolites', 'reactions',
                                                                   'interactions', 'regulators', 'targets'):

                container = obj[serialize_name]

                for var_id, variable in container.items():

                    if var_id in children:

                        continue

                    else:

                        children[var_id] = variable

        return children

    @classmethod
    def _get_container_deserializer(cls, attr):

        if attr == 'name':

            return cls._obj_deserializer

        elif attr in ('genes', 'metabolites', 'reactions',
                      'interactions', 'regulators', 'targets'):

            return cls._model_container_deserializer

        elif attr == 'objective':

            return cls._key_container_deserializer

        else:
            return lambda *args, **kwargs: {}

    @staticmethod
    def _obj_deserializer(obj, *args, **kwargs):
        return obj

    @classmethod
    def _model_container_deserializer(cls, obj, children):

        from mewpy.germ.variables import Variable

        container = {}

        for var_id, variable in obj.items():

            if isinstance(variable, Variable):

                container[var_id] = variable

            else:

                variable_dict = variable

                variable_obj = children[var_id]

                variable_attributes = cls._variable_attributes(variable_dict, variable=variable_obj, children=children)

                variable_obj.update(**variable_attributes)

                container[var_id] = variable_obj

        return container

    @staticmethod
    def _key_container_deserializer(obj, children=None, children_types=None):

        if children is False:
            return {}

        elif children is None:

            if children_types:
                from mewpy.germ.variables import Variable
                return {Variable.from_types(children_types, identifier=variable): val
                        for variable, val in obj.items()}

            else:
                from mewpy.germ.variables import Metabolite
                return {Metabolite(key): val for key, val in obj.items()}

        else:
            return {children[variable]: val for variable, val in obj.items()}

    @classmethod
    def _variable_deserializer(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                               obj):

        identifier = obj.get('id')
        types = obj.get('types')

        variable = cls.from_types(types=types, identifier=identifier)

        variable_attributes = cls._variable_attributes(obj, variable=variable)

        variable.update(**variable_attributes)

        return variable

    @classmethod
    def _variable_attributes(cls, obj, variable, children=None):

        variable_attributes = {}

        # Filtering attributes that must be used by the update method. Child variables associated with this variable
        # are also collected for further building
        for attr_name, (serialize_name, deserialize_name, _) in variable.attributes.items():

            if deserialize_name is not None:
                deserializer, children_types = cls._get_attribute_deserializer(attr=deserialize_name)

                attribute = obj[serialize_name]
                variable_attributes[deserialize_name] = deserializer(attribute,
                                                                     children=children,
                                                                     children_types=children_types)

        return variable_attributes

    @classmethod
    def _get_attribute_deserializer(cls, attr):

        if attr in ('name', 'aliases', 'coefficients', 'bounds', 'charge', 'compartment', 'formula'):

            return cls._obj_deserializer, None

        elif attr == 'regulatory_events':

            return cls._regulatory_events_deserializer, ['regulator']

        elif attr == 'gpr':

            return cls._expression_deserializer, ['gene']

        elif attr == 'reactions':

            return cls._variable_container_deserializer, ['reaction']

        elif attr == 'interactions':

            return cls._variable_container_deserializer, ['interaction']

        elif attr == 'stoichiometry':

            return cls._key_container_deserializer, ['metabolite']

        elif attr == 'target':

            return cls._variable_attribute_deserializer, ['target']

        elif attr == 'interaction':

            return cls._variable_attribute_deserializer, ['interaction']

        else:
            return lambda *args, **kwargs: {}, None

    @staticmethod
    def _expression_deserializer(obj, children=None, children_types=None):

        if children is None:

            symbolic = parse_expression(obj)

            from mewpy.germ.variables import Variable

            variables = {symbol.name: Variable.from_types(children_types, identifier=symbol.name)
                         for symbol in symbolic.atoms(symbols_only=True)}

            return Expression(symbolic=symbolic, variables=variables)

        else:

            symbolic = parse_expression(obj)

            variables = {symbol.name: children[symbol.name]
                         for symbol in symbolic.atoms(symbols_only=True)}

            return Expression(symbolic=symbolic, variables=variables)

    @staticmethod
    def _regulatory_events_deserializer(obj, children=None, children_types=None):

        return {state: Serializer._expression_deserializer(expression, children, children_types)
                for state, expression in obj.items()}

    @staticmethod
    def _variable_container_deserializer(obj, children=None, children_types=None):
        if children is None:

            from mewpy.germ.variables import Variable
            return {key: Variable.from_types(children_types, identifier=key) for key in obj}

        else:
            return {key: children[key] for key in obj}

    @staticmethod
    def _variable_attribute_deserializer(obj, children=None, children_types=None):
        if children is None:

            from mewpy.germ.variables import Variable
            return Variable.from_types(children_types, identifier=obj)

        else:
            return children[obj]

    # -----------------------------------------------------------------------------
    # reduce for pickle serialization
    # -----------------------------------------------------------------------------
    def __reduce__(self: Union['Serializer', 'Model', 'Variable']):
        # for further detail: https://docs.python.org/3/library/pickle.html#object.__reduce__

        from mewpy.germ.models import Model, build_model
        from mewpy.germ.variables import Variable, build_variable

        if isinstance(self, Model):
            return build_model, (tuple(self.types), {'identifier': self.id}), self._dict_to_pickle()

        if isinstance(self, Variable):
            return build_variable, (tuple(self.types), {'identifier': self.id}), self._dict_to_pickle()

        return super(Serializer, self).__reduce__()

    # -----------------------------------------------------------------------------
    # State for pickle serialization
    # -----------------------------------------------------------------------------

    def __getstate__(self):
        return self._dict_to_pickle()

    def __setstate__(self, state):
        self.__dict__.update(state)

    # -----------------------------------------------------------------------------
    # Pickle-like serialization
    # -----------------------------------------------------------------------------

    def _pickle_variable_serializer(self: Union['Serializer', 'Variable'],
                                    to_state=True):
        attributes = {}

        for attribute, (_, _, pickle_name) in self.attributes.items():

            if pickle_name is not None:
                attribute_obj = getattr(self, pickle_name)

                if to_state:
                    attributes[pickle_name] = attribute_obj
                else:
                    attributes[attribute] = attribute_obj

        return attributes

    def _pickle_model_serializer(self: Union['Serializer', 'Model'],
                                 to_state=True):
        containers = {}

        for container, (_, _, pickle_name) in self.containers.items():

            if pickle_name is not None:
                container_obj = getattr(self, pickle_name)

                if to_state:
                    containers[pickle_name] = container_obj
                else:
                    containers[container] = container_obj

        return containers

    def _dict_to_pickle(self: Union['Serializer', 'Model', 'Variable']):
        if hasattr(self, 'containers'):
            return self._pickle_model_serializer(to_state=True)

        if hasattr(self, 'attributes'):
            return self._pickle_variable_serializer(to_state=True)

        return {}

    @classmethod
    def _pickle_variable_deserializer(cls: Union[Type['Variable']],
                                      obj):
        identifier = obj.get('id')
        types = obj.get('types')

        variable = cls.from_types(types=types, identifier=identifier)

        new_dict = {}

        for attribute, (_, _, pickle_name) in variable.attributes.items():

            if pickle_name is not None:
                new_dict[pickle_name] = obj.get(attribute)

        variable.__dict__.update(new_dict)

        return variable

    @classmethod
    def _pickle_model_deserializer(cls: Union[Type['Model']],
                                   obj):
        identifier = obj.get('id')
        types = obj.get('types')

        model = cls.from_types(types=types, identifier=identifier)

        new_dict = {}

        for container, (_, _, pickle_name) in model.containers.items():

            if pickle_name is not None:
                new_dict[pickle_name] = obj.get(container)

        model.__dict__.update(new_dict)

        return model

    # FIXME: make sure variables point to the correct model
    def to_dict(self: Union['Serializer', 'Variable', 'Model'],
                serialization_format: str = 'json',
                variables: bool = False) -> Dict[str, Union[dict,
                                                            'Gene',
                                                            'Interaction',
                                                            'Metabolite',
                                                            'Reaction',
                                                            'Regulator',
                                                            'Target']]:
        """
        It is possible to export a variable or a model to a dictionary.
        The object can be exported to a json or a pickle format. The json format is the default.
        The json format contains the conversion of all objects to Python primitive types.
        The pickle format contains the conversion to binary code.
        To export model and variables together set variables to True.

        :param serialization_format: the format of the exported dictionary. It can be either 'json' or 'pickle'.
        :param variables: if True, the variables are exported as well. Otherwise, only the model is exported.
        :return: a dictionary containing the model or the variable.
        """

        is_model = False
        if hasattr(self, 'containers'):
            is_model = True

        if serialization_format == 'pickle':

            if is_model:
                dict_obj = self._pickle_model_serializer(to_state=False)
            else:
                dict_obj = self._pickle_variable_serializer(to_state=False)

            dict_obj['types'] = self.types
            return dict_obj

        elif serialization_format == 'json':
            if is_model:
                return self._model_serializer(variables)

            return self._variable_serializer()

        raise ValueError('The serialization format must be either json or pickle.')

    @classmethod
    def from_dict(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                  obj: Dict[str, Union[dict,
                                       'Gene',
                                       'Interaction',
                                       'Metabolite',
                                       'Reaction',
                                       'Regulator',
                                       'Target']],
                  serialization_format: str = 'json',
                  variables: bool = False) -> Union['Gene',
                                                    'Interaction',
                                                    'Metabolite',
                                                    'Reaction',
                                                    'Regulator',
                                                    'Target',
                                                    'MetabolicModel',
                                                    'RegulatoryModel']:
        """
        It is possible to create a new variable or model from a dictionary.
        The dictionary must be in the same format as the one returned by the to_dict method.
        That is, the dictionary might have the name of the containers/attributes with the primitive Python types or
        Variables objects.

        :param obj: dictionary with the information to create the variable or model
        :param serialization_format: format of the dictionary
        :param variables: if True, the dictionary must contain the information of the variables as well
        :return: a new variable or model
        """

        is_model = False
        if hasattr(cls, 'containers'):
            is_model = True

        if serialization_format == 'pickle':

            if is_model:
                return cls._pickle_model_deserializer(obj)

            return cls._pickle_variable_deserializer(obj)

        elif serialization_format == 'json':

            if is_model:
                return cls._model_deserializer(obj, variables)

            return cls._variable_deserializer(obj)

        raise ValueError('The serialization format must be either json or pickle')

    # -----------------------------------------------------------------------------
    # Copy leveraging serialization
    # -----------------------------------------------------------------------------

    def __copy__(self) -> Union['Gene',
                                'Interaction',
                                'Metabolite',
                                'Reaction',
                                'Regulator',
                                'Target',
                                'MetabolicModel',
                                'RegulatoryModel']:

        obj_dict = self.to_dict(serialization_format='pickle')

        return self.from_dict(obj_dict, serialization_format='pickle')

    def copy(self) -> Union['Gene',
                            'Interaction',
                            'Metabolite',
                            'Reaction',
                            'Regulator',
                            'Target',
                            'MetabolicModel',
                            'RegulatoryModel']:
        """
        It creates a copy of the variable or model. This is a shallow copy, meaning that the attributes and containers
        are copied by reference. If you want to create a deep copy, use the deepcopy method.
        :return: a new variable or model
        """
        return self.__copy__()

    def __deepcopy__(self, memo) -> Union['Gene',
                                          'Interaction',
                                          'Metabolite',
                                          'Reaction',
                                          'Regulator',
                                          'Target',
                                          'MetabolicModel',
                                          'RegulatoryModel']:
        obj_dict = self.to_dict(variables=True)
        memo[id(self)] = obj_dict
        return self.from_dict(obj_dict, variables=True)

    def deepcopy(self) -> Union['Gene',
                                'Interaction',
                                'Metabolite',
                                'Reaction',
                                'Regulator',
                                'Target',
                                'MetabolicModel',
                                'RegulatoryModel']:
        """
        It creates a deep copy of the variable or model.
        This means that the attributes and containers are copied by value.
        If you want to create a shallow copy, use the copy method.
        :return: a new variable or model
        """
        return self.__deepcopy__({})



from collections.abc import Callable
from typing import TYPE_CHECKING, Union
from functools import partial, wraps

import pandas as pd

if TYPE_CHECKING:
    from mewpy.germ.models import Model
    from mewpy.germ.variables import Variable


class HistoryManager:

    def __init__(self):

        self._history = []
        self._undo_able_commands = []
        self._temp_stack = []
        self._redo_able_commands = []

    def __str__(self):
        return f'History: {len(self._undo_able_commands)} undos and {len(self._redo_able_commands)} redos'

    @property
    def history(self):

        return pd.DataFrame(data=self._history, columns=['method', 'args', 'kwargs', 'object'])

    @property
    def undo_able_commands(self):
        return self._undo_able_commands

    @property
    def redo_able_commands(self):
        return self._redo_able_commands

    def _do(self, undo=True):

        if undo:
            method = self.undo_able_commands.pop()
            redo_command = self._temp_stack.pop()
            self.redo_able_commands.append(redo_command)

        else:
            method = self.redo_able_commands.pop()

        method()

    def undo(self) -> None:
        self._do(undo=True)

    def redo(self) -> None:
        self._do(undo=False)

    def reset(self) -> None:

        while len(self.undo_able_commands) > 0:
            self.undo()

    def restore(self) -> None:

        while len(self.redo_able_commands) > 0:
            self.redo()

    def __call__(self, *args, **kwargs) -> None:

        return self.queue_command(*args, **kwargs)

    def queue_command(self,
                      undo_func: Callable,
                      func: Callable,
                      undo_args: tuple = None,
                      undo_kwargs: dict = None,
                      args: tuple = None,
                      kwargs: dict = None,
                      obj: 'Model' = None) -> None:

        if not undo_args:
            undo_args = ()

        if not undo_kwargs:
            undo_kwargs = {}

        self.undo_able_commands.append(partial(undo_func, *undo_args, **undo_kwargs))

        if not args:
            args = ()

        if not kwargs:
            kwargs = {}

        self._temp_stack.append(partial(func, *args, **kwargs))

        self._history.append((func.__name__, str(args), str(kwargs), str(obj)))


def recorder(func: Callable):

    @wraps(func)
    def wrapper(self: Union['Model', 'Variable'], value):

        history = self.history

        old_value = getattr(self, func.__name__)

        if old_value != value:

            history.queue_command(undo_func=func,
                                  undo_args=(self, old_value),
                                  func=func,
                                  args=(self, value))

        func(self, value)

    return wrapper



import abc
from typing import List, Any, Dict, Callable, Union, Tuple

from .algebra_utils import solution_decode, _walk


class Symbolic:

    def __init__(self, value=None, variables=None):

        """
        The Symbolic object implements the base logic for symbolic algebra manipulation and evaluation.
        Symbolic is the base class that provides the abstraction and tools for a given variable or operator be handled
        as an algebra operand or operator, respectively.

        Operators and Operands that inherit from Symbolic implement python dunder methods (e.g. __add__), so that they
        can be used in python evaluation method (eval) during binary string evaluation

        Symbolic objects can represent either atoms (e.g. Symbol) or nodes (e.g. And)
        of a binary Abstract Syntax Tree (AST). Thus, each symbolic-object can hold a given value (e.g. Symbol('x'))
        or hold variables/children (e.g. And(variables=[Symbol('x'), Symbol('y')])).

        The symbolic object at the root of the algebra expression (AST) holds the variables and subsequent descents.

        A symbolic object is usually assembled during parsing of an algebra expression encoded into a string object.

        :param value: The value that the symbolic operand is representing in the algebra expression.
        It can either be a numeric or variable value
        :param variables: The variables/children that the symbolic operator is holding in the algebra expression.
        It can either be a Symbol, Numeric, Boolean, or other atoms.
        """

        if not variables:
            variables = []

        self.value = value
        self.variables = variables

        self._name = None
        self.model_variable = None

        self.is_boolean = False
        self.is_true = False
        self.is_false = False
        self.is_and = False
        self.is_or = False
        self.is_not = False

        self.is_relational = False
        self.is_equal = False
        self.is_not_equal = False
        self.is_inequality = False
        self.is_greater = False
        self.is_greater_equal = False
        self.is_less = False
        self.is_less_equal = False

        self.is_numeric = False
        self.is_integer = False
        self.is_float = False
        self.is_one = False
        self.is_zero = False

        self.is_symbol = False
        self.is_atom = False

        self.is_none = False

    @property
    def name(self) -> str:
        """
        String representation of the symbolic object value
        """

        if self._name is None:
            return str(self.value)

        return self._name

    @name.setter
    def name(self, value):

        if value is None:
            return

        self._name = value

    @property
    def bounds(self) -> Tuple[float, float]:
        """
        Returns the bounds of the symbolic object
        """
        if self.model_variable:
            if hasattr(self.model_variable, 'bounds'):
                return self.model_variable.bounds

            if hasattr(self.model_variable, 'coefficients'):
                return min(self.model_variable.coefficients), max(self.model_variable.coefficients)

        return 0, 1

    def key(self):
        return self.to_string().replace(' ', '')

    def __repr__(self):

        if self.is_atom:

            return f"{self.__class__.__name__}({self.value})"

        else:

            return f"{self.__class__.__name__}(variables={self.variables})"

    def __str__(self):

        string_repr = ''

        if self.is_and:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' & '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_or:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' | '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_not:
            string_repr += '('
            for child in self.variables:
                string_repr += ' ~ ' + child.to_string()
            string_repr += ')'

        elif self.is_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' = '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_not_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' != '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_inequality:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' > '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_greater:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' > '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_greater_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' >= '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_less:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' < '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_less_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' <= '
            string_repr = string_repr[:-3]
            string_repr += ')'

        else:

            string_repr += self.name

        return string_repr

    def to_string(self) -> str:
        """
        String representation of this Symbolic object
        """
        return self.__str__()

    def __iter__(self):

        self.__variables_iter = list(_walk(self, reverse=True))
        self.__i = -1
        self.__i_limit = len(self.__variables_iter)

        return self

    def __next__(self):

        self.__i += 1

        if self.__i < self.__i_limit:
            return self.__variables_iter[self.__i]

        raise StopIteration

    def atoms(self, symbols_only=False) -> List['Symbolic']:
        """
        It returns all symbolic atoms associated with this symbolic object

        :param symbols_only: Whether to return only symbols or not
        :return: It returns a list of symbolic atoms
        """

        symbols = list(_walk(self))

        if symbols_only:

            return [symbol for symbol in symbols if symbol.is_symbol]

        else:

            return [symbol for symbol in symbols if symbol.is_atom]

    def _tree_evaluation(self, values, operators, default, **kwargs):

        left = None
        right = None

        if not self.variables:
            return self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        # noinspection PyProtectedMember
        left = self.variables[0]._tree_evaluation(values=values, operators=operators, default=default, **kwargs)

        for child in self.variables[1:]:
            # noinspection PyProtectedMember
            right = child._tree_evaluation(values=values, operators=operators, default=default, **kwargs)
            left = self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        # not special case
        if right is None:
            left = self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        return left

    def evaluate(self,
                 values: Dict[str, Any] = None,
                 operators: Dict[str, Callable] = None,
                 default=0,
                 **kwargs) -> Union[float, int, Any]:

        """
        Evaluate a Symbolic algebra expression based on the coefficients/values of the symbols or atoms contained
        in the expression.

        The symbolic expression can be evaluated according to the custom values
        assigned to the symbols in the values' dictionary. Likewise, operators evaluation can also be altered
        by passing a callable to a given operator in the operators' dictionary.

        A symbol is also a callable.

        :param values: A dictionary of values that the symbols names must take during expression evaluation
        :param operators: A dictionary of custom operators. That is, python operators-based evaluation
        (e.g. 3 > 2 yield True) can be replaced by custom callable objects such as functions. For instance,
        3 > 2 can be evaluated with max, and thus max(3, 2) yields 3 now.
        :param default: The default value of a given symbol/variable.
        The default value is used if a given variable/symbol is missing in the values dictionary.
        :param kwargs: Additional keyword arguments for Symbolic evaluate method

        :return: The solution of the Symbolic expression evaluation as int, float or Any type.
        """

        if not values:
            values = {}

        if not operators:
            operators = {}

        return self._tree_evaluation(values=values, operators=operators, default=default, **kwargs)

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return

    def __call__(self, values=None, operators=None, default=0, **kwargs):

        return self.evaluate(values=values,
                             operators=operators,
                             default=default,
                             **kwargs)


class Boolean(Symbolic):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_boolean = True

    def __and__(self, *args):
        return And(variables=[self, *args])

    __rand__ = __and__

    def __or__(self, *args):
        return Or(variables=[self, *args])

    __ror__ = __or__

    def __invert__(self):
        return Not(variables=[self])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class BoolFalse(Boolean):
    """
    Atomic Boolean False value
    """

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 0

        self.is_false = True
        self.is_atom = True

    def __bool__(self):
        return False

    @staticmethod
    def _evaluate(**kwargs):
        return 0


class BoolTrue(Boolean):
    """
    Atomic Boolean True value
    """

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 1

        self.is_true = True
        self.is_atom = True

    def __bool__(self):
        return True

    @staticmethod
    def _evaluate(**kwargs):
        return 1


class And(Boolean):
    """
    Boolean Operator AND
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_and = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(And, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(int(left) & int(right))

    def __and__(self, *args):
        self.variables.extend(args)
        return self

    __rand__ = __and__


class Or(Boolean):
    """
    Boolean Operator OR
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_or = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Or, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(int(left) | int(right))

    def __or__(self, *args):
        self.variables.extend(args)
        return self

    __ror__ = __or__


class Not(Boolean):
    """
    Boolean Operator NOT
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_not = True

    @staticmethod
    def _evaluate(left=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Not, None)

        if operator is not None:
            return operator(left)

        return solution_decode(~ int(left))


class Relational(Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_relational = True

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Equal(Relational):
    """
    Relational Operator EQUAL
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Equal, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left == right)


class NotEqual(Relational):
    """
    Relational Operator NOT EQUAL
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_not_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(NotEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left != right)


class Inequality(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_inequality = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Inequality, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left > right)


class Greater(Relational):
    """
    Relational Operator Greater
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_greater = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Greater, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left > right)


class GreaterEqual(Relational):
    """
    Relational Operator Greater Equal
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_greater_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(GreaterEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left >= right)


class Less(Relational):
    """
    Relational Operator Less
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_less = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Less, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left < right)


class LessEqual(Relational):
    """
    Relational Operator Less Equal
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_less_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(LessEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left <= right)


class AtomNumber(Symbolic):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_numeric = True
        self.is_atom = True

    def __eq__(self, other):
        return Equal(variables=[self, other])

    def __ne__(self, other):
        return NotEqual(variables=[self, other])

    def __ge__(self, other):
        return GreaterEqual(variables=[self, other])

    def __le__(self, other):
        return LessEqual(variables=[self, other])

    def __gt__(self, other):
        return Greater(variables=[self, other])

    def __lt__(self, other):
        return Less(variables=[self, other])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Number(AtomNumber):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __eq__(self, other):
        return Equal(variables=[self, other])

    def __ne__(self, other):
        return NotEqual(variables=[self, other])

    def __ge__(self, other):
        return GreaterEqual(variables=[self, other])

    def __le__(self, other):
        return LessEqual(variables=[self, other])

    def __gt__(self, other):
        return Greater(variables=[self, other])

    def __lt__(self, other):
        return Less(variables=[self, other])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Integer(Number):
    """
    Atomic Numeric value Integer
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = int(self.value)

        self.is_integer = True

    def _evaluate(self, **kwargs):
        return self.value


class Float(Number):
    """
    Atomic Numeric value Float
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = float(self.value)

        self.is_float = True

    def _evaluate(self, **kwargs):
        return self.value


class Zero(Boolean, AtomNumber):
    """
    Atomic Numeric value Zero
    """

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 0

        self.is_zero = True

    def __bool__(self):
        return False

    def _evaluate(self, **kwargs):
        return 0


class One(Boolean, AtomNumber):
    """
    Atomic Numeric value One
    """

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 1

        self.is_one = True

    def __bool__(self):
        return True

    def _evaluate(self, **kwargs):
        return 1


class Symbol(AtomNumber, Boolean):
    """
    Atomic Symbolic Variable
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_boolean = False
        self.is_numeric = False

        self.is_symbol = True

    def _evaluate(self, values=None, default=0, **kwargs):
        if values is None:
            values = {}

        if self.name in values:
            value = values[self.name]

        elif self.model_variable:
            if hasattr(self.model_variable, 'bounds'):
                value = max(self.model_variable.bounds)

            elif hasattr(self.model_variable, 'coefficients'):
                value = max(self.model_variable.coefficients)

            else:
                value = default

        else:
            value = default

        if not isinstance(value, (int, float, bool, AtomNumber)):
            raise ValueError(f'cannot evaluate {self.name} for {value}')

        return solution_decode(value)


class NoneAtom(Symbol):
    """
    Atomic Symbolic Empty value
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_none = True
        self.is_symbol = False

    @property
    def name(self):
        return ''

    def _evaluate(self, **kwargs):
        return 0



from typing import Union, Dict

from mewpy.germ.analysis import FBA
from mewpy.germ.lp import ConstraintContainer, VariableContainer
from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
from mewpy.solvers.solution import Solution, Status
from mewpy.solvers.solver import VarType, Solver


class pFBA(FBA):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        Parsimonious Flux Balance Analysis (FBA) of a metabolic model.
        Regular implementation of a pFBA for a metabolic model.

        This pFBA implementation was heavily inspired by pFBA implementation of reframed python package. Take a look at
        the source: https://github.com/cdanielmachado/reframed and https://reframed.readthedocs.io/en/latest/

        For more details consult: https://doi.org/10.1038/msb.2010.47

        :param model: a MetabolicModel, RegulatoryModel or GERM model. The model is used to retrieve
        variables and constraints to the linear problem
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used, but it will be overwritten
        if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        super().__init__(model=model, solver=solver, build=build, attach=attach)

    def _wt_bounds(self, fraction: float = None, solver_kwargs: Dict = None):
        """
        It builds the linear problem from the model. The linear problem is built from the model
        variables and constraints. The linear problem is then loaded into the solver.
        :return:
        """
        if not solver_kwargs:
            solver_kwargs = {}

        sol = FBA(model=self.model, build=True, attach=False).optimize(solver_kwargs=solver_kwargs, to_solver=True)
        if sol.status != Status.OPTIMAL:
            lb, ub = 0.0, 0.0

        else:
            if fraction is None:
                lb, ub = float(sol.fobj), float(sol.fobj)
            else:
                lb, ub = float(sol.fobj) * fraction, float(sol.fobj)
        return lb, ub

    def _build_pfba_constrains(self, fraction: float = None, solver_kwargs: Dict = None):
        """
        It builds the pfba constraints of the linear problem.
        :return:
        """
        if not solver_kwargs:
            solver_kwargs = {}

        lb, ub = self._wt_bounds(fraction, solver_kwargs)

        if 'linear' in solver_kwargs:
            coef = solver_kwargs['linear'].copy()
        else:
            coef = {variable.id: val for variable, val in self.model.objective.items()}

        if 'constraints' in solver_kwargs:
            constraints = solver_kwargs['constraints'].copy()
        else:
            constraints = {}

        constraint = ConstraintContainer(name='pfba_constraints', coefs=[coef], lbs=[lb], ubs=[ub])
        variable = VariableContainer(name='pfba_variables', sub_variables=[], lbs=[], ubs=[], variables_type=[])
        objective = {}
        for reaction in self.model.yield_reactions():

            if reaction.reversibility:
                rxn_forward = f'{reaction.id}_forward'
                rxn_reverse = f'{reaction.id}_reverse'

                rxn_ub = float(constraints.get(reaction.id, reaction.bounds)[1])

                variable.sub_variables.extend([rxn_forward, rxn_reverse])
                variable.lbs.extend([0.0, 0.0])
                variable.ubs.extend([rxn_ub, rxn_ub])
                variable.variables_type.extend([VarType.CONTINUOUS, VarType.CONTINUOUS])

                constraint.lbs.extend([0.0, 0.0])
                constraint.ubs.extend([rxn_ub, rxn_ub])
                constraint.coefs.extend([{reaction.id: -1, rxn_forward: 1},
                                         {reaction.id: 1, rxn_reverse: 1}])

                objective[rxn_forward] = 1
                objective[rxn_reverse] = 1

            else:
                objective[reaction.id] = 1

        self.add_variables(variable)
        self.add_constraints(constraint)
        self._linear_objective = objective
        self._minimize = True

    def _build(self):
        """
        It builds the linear problem from the model. The linear problem is built from the model
        variables and constraints. The linear problem is then loaded into the solver.
        :return:
        """
        if self.model.is_metabolic():
            # mass balance constraints and reactions' variables
            self._build_mass_constraints()

            # pFBA constraints can be added again to the linear problem during optimization
            self._build_pfba_constrains()

        return

    def _optimize(self, fraction: float = None, solver_kwargs: Dict = None, **kwargs) -> Solution:
        """
        It optimizes the linear problem. The linear problem is solved by the solver interface.
        :param solver_kwargs: A dictionary of keyword arguments to be passed to the solver.
        :return: A Solution instance.
        """
        if not solver_kwargs:
            solver_kwargs = {}

        linear = solver_kwargs.get('linear')
        constraints = solver_kwargs.get('constraints')

        # if linear and constraints are not provided, build new pfba constraints and solver
        replace_pfba_constraints = [x for x in (fraction, linear, constraints) if x is not None]

        if replace_pfba_constraints:
            self._build_pfba_constrains(solver_kwargs=solver_kwargs)
            self.build_solver()

        solution = self.solver.solve(**solver_kwargs)

        # restore the pfba constraints and solver to the previous state
        if replace_pfba_constraints:
            self._build_pfba_constrains()
            self.build_solver()

        return solution



from typing import Union, Dict, Tuple, List, TYPE_CHECKING, Optional

import logging
import numpy as np
import pandas as pd

from . import SimulationMethod, SStatus
from .simulation import Simulator, SimulationResult, ModelContainer
from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
from mewpy.germ.variables import Reaction
from mewpy.util.constants import ModelConstants
from mewpy.util.utilities import Dispatcher, AttrDict
from mewpy.germ.analysis import FBA, pFBA, fva
from mewpy.solvers.solution import Solution, Status
from tqdm import tqdm

if TYPE_CHECKING:
    from mewpy.solvers.solver import Solver
    from mewpy.solvers.cplex_solver import CplexSolver
    from mewpy.solvers.gurobi_solver import GurobiSolver
    from mewpy.solvers.optlang_solver import OptLangSolver

LOGGER = logging.getLogger(__name__)


class GERMModel(ModelContainer):

    def __init__(self, model: Union[Model, MetabolicModel, RegulatoryModel]):
        """
        A GERM model interface to access GERM model containers

        :type model: Union[Model, MetabolicModel, RegulatoryModel]
        :param model: A GERM Model, MetabolicModel, RegulatoryModel or all together
        """

        if not isinstance(model, (Model, MetabolicModel, RegulatoryModel)):
            raise ValueError("The model is not an instance of mewpy.germ.model.Model")

        self.model = model

    # -----------------------------------------------------------------------------
    # Metabolic static attributes
    # -----------------------------------------------------------------------------
    @property
    def reactions(self) -> List[str]:
        """
        Returns the reactions identifiers
        :return: a list of reactions identifiers
        """
        if self.model.is_metabolic():
            return list(self.model.reactions.keys())
        return []

    @property
    def metabolites(self) -> List[str]:
        """
        Returns the metabolites identifiers
        :return: a list of metabolites identifiers
        """
        if self.model.is_metabolic():
            return list(self.model.metabolites.keys())
        return []

    @property
    def genes(self) -> List[str]:
        """
        Returns the genes identifiers
        :return: a list of genes identifiers
        """
        if self.model.is_metabolic():
            return list(self.model.genes.keys())
        return []

    @property
    def proteins(self) -> List[str]:
        """
        GERM models do not support proteins yet
        :return: empty list
        """
        return []

    @property
    def medium(self) -> Dict[str, float]:
        """
        Returns the medium
        :return: a dictionary of exchange reaction identifiers and lower bounds
        """
        return {rxn.id: rxn.lower_bound for rxn in self.model.yield_exchanges()
                if rxn.lower_bound < ModelConstants.TOLERANCE}

    # -----------------------------------------------------------------------------
    # Metabolic dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def compartments(self) -> Dict[str, str]:
        """
        Returns the compartments
        :return: a dictionary of compartments
        """
        if self.model.is_metabolic():
            return self.model.compartments

        return {}

    def get_reaction(self, r_id: str) -> AttrDict:
        """
        It returns a reaction by its identifier
        :param r_id: reaction identifier
        :return: a reaction dict
        """
        if self.model.is_metabolic():
            reaction = self.model.get(r_id)
            if reaction:
                reaction = {
                    'id': reaction.id,
                    'name': reaction.name,
                    'lower_bound': reaction.lower_bound,
                    'upper_bound': reaction.upper_bound,
                    'stoichiometry': {met.id: c for met, c in reaction.stoichiometry.items()},
                    'gpr': reaction.gene_protein_reaction_rule,
                }
                AttrDict(reaction)
        return AttrDict()

    def get_gene(self, g_id: str) -> AttrDict:
        """
        It returns a gene by its identifier
        :param g_id: gene identifier
        :return: a gene dict
        """
        if self.model.is_metabolic():
            gene = self.model.get(g_id)
            if gene:
                gene = {'id': gene.id,
                        'name': gene.name,
                        'reactions': list(gene.reactions.keys())}
                return AttrDict(gene)
        return AttrDict()

    def get_compartment(self, c_id: str) -> AttrDict:
        """
        It returns a compartment by its identifier
        :param c_id: compartment identifier
        :return: a compartment dict
        """
        if self.model.is_metabolic():
            compartment = self.model.compartments.get(c_id)
            if compartment:
                compartment = {'id': c_id,
                               'name': compartment,
                               'external': c_id == self.model.external_compartment}
                return AttrDict(compartment)
        return AttrDict()

    def get_gene_reactions(self) -> Dict[str, List[str]]:
        """
        It returns a dictionary of genes and their associated reactions
        :return: a dictionary of genes and their associated reactions
        """
        if self.model.is_metabolic():
            return {gene.id: list(gene.reactions.keys()) for gene in self.model.yield_genes()}
        return {}

    def get_exchange_reactions(self) -> List[str]:
        """
        Returns the exchange reactions
        :return: a list of exchange reactions
        """
        if self.model.is_metabolic():
            return list(self.model.exchanges.keys())
        return []

    def get_gpr(self, reaction_id) -> Optional[str]:
        """
        Returns the gpr of a reaction
        :param reaction_id: reaction identifier
        :return: a string representation of the gpr
        """
        if self.model.is_metabolic():

            reaction = self.model.get(reaction_id)

            if reaction:

                if not reaction.gpr.is_none:
                    return reaction.gpr.to_string()

        return

    # -----------------------------------------------------------------------------
    # Metabolic static attributes setters
    # -----------------------------------------------------------------------------
    def set_objective(self, reaction):
        """
        It sets the objective function
        :param reaction: reaction identifier
        :return:
        """
        if self.model.is_metabolic():
            self.model.objective = reaction

    # -----------------------------------------------------------------------------
    # Regulatory static attributes
    # -----------------------------------------------------------------------------
    @property
    def interactions(self) -> List[str]:
        """
        It returns the interactions identifiers
        :return: a list of interactions identifiers
        """
        if self.model.is_regulatory():
            return list(self.model.interactions.keys())
        return []

    @property
    def regulators(self) -> List[str]:
        """
        It returns the regulators identifiers
        :return: a list of regulators identifiers
        """
        if self.model.is_regulatory():
            return list(self.model.regulators.keys())
        return []

    @property
    def targets(self) -> List[str]:
        """
        It returns the targets identifiers
        :return: a list of targets identifiers
        """
        if self.model.is_regulatory():
            return list(self.model.targets.keys())
        return []

    @property
    def environmental_stimuli(self) -> List[str]:
        """
        It returns the environmental stimuli identifiers
        :return: a list of environmental stimuli identifiers
        """
        if self.model.is_regulatory():
            return list(self.model.environmental_stimuli.keys())
        return []

    # -----------------------------------------------------------------------------
    # Model summary
    # -----------------------------------------------------------------------------
    def summary(self):
        """
        It prints a summary of the model
        :return:
        """
        if self.model.is_metabolic():
            print(f"Metabolites: {len(self.metabolites)}")
            print(f"Reactions: {len(self.reactions)}")
            print(f"Genes: {len(self.genes)}")

        if self.model.is_regulatory():
            print(f"Interactions: {len(self.interactions)}")
            print(f"Regulators: {len(self.regulators)}")
            print(f"Targets: {len(self.targets)}")


class Simulation(GERMModel, Simulator):
    dispatcher = Dispatcher()

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 envcond: Dict[str, Tuple[Union[int, float], Union[int, float]]] = None,
                 constraints: Dict[str, Tuple[Union[int, float], Union[int, float]]] = None,
                 reference: Dict[str, Union[int, float]] = None,
                 reset_solver=ModelConstants.RESET_SOLVER):
        """
        It supports simulation of a MetabolicModel, RegulatoryModel or GERM model.
        Additional environmental conditions and constraints can be set using this interface.
        The objective function can be altered too.
        A reference wild type reaction flux values can be provided for convenience.

        Simulation offers several methods to retrieve additional information from the model.

        :type model: Union[Model, MetabolicModel, RegulatoryModel]
        :param model: A germ Model, MetabolicModel, RegulatoryModel or all together

        Optional:

        :param envcond: a dictionary of additional environmental conditions
        :param constraints: a dictionary of additional constraints
        :param reference: A dictionary of the wild type flux values
        :param reset_solver: Whether it is required to reset the solver or not after simulation.
        """

        super().__init__(model)

        if not isinstance(model, (Model, MetabolicModel, RegulatoryModel)):
            raise ValueError("The model is not an instance of slim.core.Model")

        if not envcond:
            envcond = {}

        if not constraints:
            constraints = {}

        self.model = model

        self.environmental_conditions = envcond
        self.constraints = constraints

        self._essential_reactions = None
        self._essential_genes = None
        self._reference = reference
        self._reset_solver = reset_solver

        self._m_r_lookup = None

        self.__status_mapping = {
            Status.OPTIMAL: SStatus.OPTIMAL,
            Status.UNBOUNDED: SStatus.UNBOUNDED,
            Status.INFEASIBLE: SStatus.INFEASIBLE,
            Status.INF_OR_UNB: SStatus.INF_OR_UNB,
            Status.UNKNOWN: SStatus.UNKNOWN,
            Status.SUBOPTIMAL: SStatus.SUBOPTIMAL
        }

    # -----------------------------------------------------------------------------
    # Metabolic interface
    # -----------------------------------------------------------------------------
    @property
    def objective(self) -> Dict[str, Union[float, int]]:
        """
        It returns the objective function, namely the variable identifier and its coefficient
        :return: a dictionary of the objective function
        """
        return {var.id: coef for var, coef in self.model.objective.items()}

    @objective.setter
    def objective(self, value: Dict[str, Union[float, int]]):
        """
        It sets the objective function
        :param value: a dictionary of the objective function, namely the variable identifier and its coefficient
        :return:
        """
        self.model.objective = value

    @property
    def reference(self) -> Dict[str, Union[float, int]]:
        """
        The reference wild type reaction flux values.
        :return: A dictionary of wild type reaction flux values.
        """
        if self._reference is None:
            self._reference = self.simulate(method=SimulationMethod.pFBA).fluxes
        return self._reference

    def essential_reactions(self, min_growth: float = 0.01) -> List[str]:
        """
        The set of knocked out reactions that impair a minimal growth rate predicted by the model with fba.
        The defined minimal percentage of the wild type growth rate is usually set to 0.01 (1%).
        :return: A list of essential reactions.
        """

        if self._essential_reactions is not None:
            return self._essential_reactions

        if not self.model.is_metabolic():
            return []

        wt_solution = self.simulate(method=SimulationMethod.FBA)
        wt_growth = wt_solution.objective_value

        self._essential_reactions = []

        reactions = self.reactions.copy()
        for rxn in tqdm(reactions):

            res = self.simulate(constraints={rxn: 0})

            if res:

                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_reactions.append(rxn)

        return self._essential_reactions

    def essential_genes(self, min_growth: float = 0.01) -> List[str]:
        """
        The set of knocked out genes that impair a minimal growth rate predicted by the model with fba.
        The defined minimal percentage of the wild type growth rate is usually set to 0.01 (1%).
        :return: A list of essential genes.
        """

        if self._essential_genes is not None:
            return self._essential_genes

        if not self.model.is_metabolic():
            return []

        self._essential_genes = []

        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value

        values = {gene.id: max(gene.coefficients) for gene in self.model.yield_genes()}

        for gene in tqdm(self.model.yield_genes()):

            gene_coefficient = values.pop(gene.id)
            values[gene.id] = 0.0

            constraints = {}
            for rxn in gene.yield_reactions():

                if rxn.gpr.is_none:
                    continue

                res = rxn.gpr.evaluate(values=values)

                if not res:
                    constraints[rxn.id] = (0.0, 0.0)

            if not constraints:
                values[gene.id] = gene_coefficient
                continue

            res = self.simulate(constraints=constraints)
            if not res:
                values[gene.id] = gene_coefficient
                continue

            if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                    or res.status == SStatus.INFEASIBLE:
                self._essential_genes.append(gene)

            values[gene.id] = gene_coefficient

        return self._essential_genes

    def evaluate_gprs(self, active_genes: List[str]) -> List[str]:
        """
        Returns the list of active reactions for a given list of active genes.
        :param list active_genes: list of genes identifiers.
        :return: a list of active reaction identifiers.
        """

        if not self.model.is_metabolic():
            return []

        values = {gene.id: 1.0 if gene.id in active_genes else 0.0
                  for gene in self.model.yield_genes()}

        return [rxn.id for rxn in self.model.yield_reactions()
                if rxn.gpr.is_none or rxn.gpr.evaluate(values=values)]

    def add_reaction(self, reaction: Reaction, replace: bool = True, comprehensive: bool = True):
        """
        Adds a reaction to the germ model
        :param reaction: Reaction object to be added
        :param replace: Whether to replace the reaction in the model if already exists
        :param comprehensive: Whether to add metabolites and genes associated with the reaction to the model
        :return:
        """
        if not self.model.is_metabolic():
            return

        if replace:
            self.model.add(reaction, comprehensive=comprehensive, history=False)

        else:
            if reaction.id not in self.model.reactions:
                self.model.add(reaction, comprehensive=comprehensive, history=False)

    def remove_reaction(self, reaction: Reaction, remove_orphans: bool = True):
        """
        Removes a reaction from the germ model
        :param reaction: Reaction object to be removed
        :param remove_orphans: Whether to remove metabolites and genes orphans associated with the reaction in the model
        :return:
        """
        if not self.model.is_metabolic():
            return

        self.model.remove(reaction, remove_orphans=remove_orphans, history=False)

    def get_uptake_reactions(self) -> List[str]:
        """
        Get uptake reactions
        :return: The list of uptake reactions.
        """
        if not self.model.is_metabolic():
            return []

        return [rxn.id for rxn in self.model.yield_exchanges()
                if rxn.reversibility or rxn.lower_bound < 0]

    def get_transport_reactions(self) -> List[str]:
        """
        Get the transport reactions
        :return: The list of transport reactions.
        """
        if not self.model.is_metabolic():
            return []

        ext = self.model.external_compartment

        return [rxn.id for rxn in self.model.yield_reactions()
                if ext in rxn.compartments and len(rxn.compartments) > 1]

    def get_transport_genes(self) -> List[str]:
        """
        Get genes associated with transport reactions
        :return: the list genes associated with transport reactions.
        """
        if not self.model.is_metabolic():
            return []

        ext = self.model.external_compartment

        genes = set()

        for rxn in self.model.yield_reactions():

            if ext in rxn.compartments and len(rxn.compartments) > 1:
                genes.update(rxn.genes.keys())

        return list(genes)

    def reverse_reaction(self, reaction_id: str) -> Union[str, None]:
        """
        Identify if a reaction is reversible and returns the reverse reaction if it is the case.

        :param reaction_id: a reaction identifier
        :return: a reverse reaction identifier or None
        """
        reaction = self.model.get(reaction_id)
        if reaction:
            if reaction.reversibility:
                return reaction_id
        return

    def metabolite_reaction_lookup(self, force_recalculate: bool = False) -> Dict[str, Dict[str, Union[int, float]]]:
        """
        A view of the S matrix for the model.
        That is, it provides the network topology as a nested map from metabolite to reaction to coefficient.
        :return: a dictionary lookup table
        """
        if not self._m_r_lookup or force_recalculate:

            metabolites = self.metabolites.copy()

            self._m_r_lookup = {met: {} for met in metabolites}

            for reaction in self.model.yield_reactions():

                for met, coef in reaction.stoichiometry.items():
                    self._m_r_lookup[met.id][reaction.id] = coef

        return self._m_r_lookup

    def get_reaction_bounds(self, reaction: str) -> Tuple[Union[int, float], Union[int, float]]:
        """
        Get the bounds for a given reaction.
        :param reaction: reaction identifier
        :return: a tuple with the reaction bounds
        """
        if reaction in self.constraints:
            lb, ub = self.constraints[reaction]

        elif reaction in self.environmental_conditions:
            lb, ub = self.environmental_conditions[reaction]

        else:
            lb, ub = self.model.get(reaction).bounds

        return lb if lb > -np.inf else -999999, ub if ub < np.inf else 999999

    def find_bounds(self) -> Tuple[Union[int, float], Union[int, float]]:
        """
        It finds the median upper and lower bound of the metabolic model.
        Bounds can vary from model to model. germ set reactions bounds to (-999999.0, 999999.0) by default.

        :return: a tuple of median upper and lower bounds
        """
        lbs, ubs = [], []

        for rxn in self.model.yield_reactions():
            lbs.append(rxn.lower_bound)
            ubs.append(rxn.upper_bound)

        lbs = np.asarray(lbs, dtype=float)
        ubs = np.asarray(ubs, dtype=float)

        lb = np.nanmedian(lbs[lbs != 0.0])
        ub = np.nanmedian(ubs[ubs != 0.0])

        if np.isnan(lb):
            LOGGER.warning("Could not identify a median lower bound. Setting the default")
            lb = -1000.0

        if np.isnan(ub):
            LOGGER.warning("Could not identify a median upper bound. Setting the default")
            ub = 1000.0

        return lb, ub

    def find_unconstrained_reactions(self) -> List[str]:
        """
        Finds a list of reactions that are not constrained in the model
        :return: list of unconstrained reactions
        """
        lb, ub = self.find_bounds()

        return [rxn.id for rxn in self.model.yield_reactions()
                if rxn.lower_bound <= lb and rxn.upper_bound >= ub]

    # -----------------------------------------------------------------------------
    # Simulation
    # -----------------------------------------------------------------------------
    @dispatcher.register(SimulationMethod.FBA)
    def _fba(self, model, objective, minimize, constraints, *args, **kwargs):
        fba = FBA(model).build()

        solver_kwargs = {'linear': objective, 'minimize': minimize, 'constraints': constraints}
        sol = fba.optimize(solver_kwargs=solver_kwargs, to_solver=True, get_values=True)
        return sol

    @dispatcher.register(SimulationMethod.pFBA)
    def _pfba(self, model, objective, minimize, constraints, *args, **kwargs):
        pfba = pFBA(model).build()

        solver_kwargs = {'linear': objective, 'minimize': minimize, 'constraints': constraints}
        sol = pfba.optimize(solver_kwargs=solver_kwargs, to_solver=True, get_values=True)
        return sol

    @dispatcher.register(SimulationMethod.MOMA)
    def _moma(self, *args, **kwargs):

        raise NotImplementedError('MOMA is not currently available with GERM models')

    @dispatcher.register(SimulationMethod.lMOMA)
    def _lmoma(self, *args, **kwargs):

        raise NotImplementedError('lMOMA is not currently available with GERM models')

    @dispatcher.register(SimulationMethod.ROOM)
    def _romm(self, *args, **kwargs):

        raise NotImplementedError('ROOM is not currently available with GERM models')

    @dispatcher.register(SimulationMethod.NONE)
    def _none(self, *args, **kwargs):
        return Solution()

    def simulate(self,
                 objective: Dict[str, Union[int, float]] = None,
                 method: SimulationMethod = SimulationMethod.FBA,
                 maximize: bool = True,
                 constraints: Union[Dict[str, float], Dict[str, Tuple[float, float]]] = None,
                 reference: Dict[str, Union[int, float]] = None,
                 scalefactor: float = None,
                 solver: Union['Solver', 'CplexSolver', 'GurobiSolver', 'OptLangSolver'] = None) -> SimulationResult:
        """
        Simulates a phenotype for a given objective and set of constraints using the specified method.
        Reference wild-type conditions are also accepted

        :param objective: a dictionary with the objective function
        :param method: a simulation method
        :param maximize: maximize or minimize the objective function
        :param constraints: a dictionary with additional constraints used only in this simulation
        :param reference: a dictionary with the reference wild-type conditions for KO simulations
        :param scalefactor: a scale factor for the reference wild-type conditions
        :param solver: a solver instance
        :return: SimulationResult object that holds the model, objective, solution values, status,
        environmental conditions and further constraints
        """
        if not objective:

            objective = {}

            if self.model.is_metabolic():
                objective = {var.id: val for var, val in self.model.objective.items()}

        if not constraints:
            constraints = {}

        simulation_constraints = {**constraints, **self.constraints, **self.environmental_conditions}

        solution: Solution = self.dispatcher(method,
                                             model=self.model,
                                             objective=objective,
                                             minimize=not maximize,
                                             constraints=simulation_constraints)

        status = self.__status_mapping[solution.status]

        return SimulationResult(model=self.model,
                                objective_value=solution.fobj,
                                fluxes=solution.values,
                                status=status,
                                envcond=self.environmental_conditions,
                                model_constraints=self.constraints,
                                simul_constraints=constraints,
                                maximize=maximize,
                                method=method)

    def FVA(self,
            obj_frac: float = 0.9,
            reactions: List[str] = None,
            constraints: Union[Dict[str, float], Dict[str, Tuple[float, float]]] = None,
            loopless: bool = False,
            internal: List[str] = None,
            solver: Union['Solver', 'CplexSolver', 'GurobiSolver', 'OptLangSolver'] = None,
            format: str = 'dict'):
        """
        It performs a Flux Variability Analysis (FVA).

        :param obj_frac: The minimum fraction of the maximum growth rate (default 0.9).
        Requires that the objective value is at least the fraction times maximum objective value.
        A value of 0.85 for instance means that the objective has to be at least at 85% percent of its maximum.
        :param reactions: List of reactions to analyze (default: all).
        :param constraints: Additional constraints (optional).
        :param loopless: Run looplessFBA internally (very slow) (default: false).
        :param internal: List of internal reactions for looplessFBA (optional).
        :param solver: A pre-instantiated solver instance (optional)
        :param format: The return format: 'dict' to return a dictionary; 'df' to return a data frame.

        :return: A dictionary or data frame of flux variation ranges.
        """

        if not constraints:
            constraints = {}

        simulation_constraints = {**constraints, **self.constraints, **self.environmental_conditions}

        solution = fva(model=self.model,
                       fraction=obj_frac,
                       reactions=reactions,
                       constraints=simulation_constraints)

        if format == 'df':
            df = pd.concat([pd.DataFrame(solution.index), solution], axis=1)
            df.columns = ['Reaction ID', 'Minimum', 'Maximum']

            return df

        return solution.to_dict()



from abc import ABCMeta, abstractmethod
from collections import defaultdict

from typing import Union, TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from mewpy.germ.models import RegulatoryModel, MetabolicModel, Model
    from mewpy.io.dto import DataTransferObject
    from io import TextIOWrapper
    from cobra import Model as CobraModel
    from reframed import CBModel as ReframedModel


class Engine(metaclass=ABCMeta):

    def __init__(self,
                 io: Union[str, 'TextIOWrapper', 'CobraModel', 'ReframedModel'],
                 config: dict,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):
        """
        The engine interface for reading/writing models. The abstract properties and methods should be fully
        implemented in the concrete engines.

        The concrete engines will open, parse and read specific types to a GERM model,
        such as:
            - regulatory sbml
            - metabolic sbml
            - regulatory csv
            - cbm models objects from external packages
            - json files

        The concrete engines will also open and write specific types from an adequate GERM model,
        such as:
            - regulatory sbml
            - metabolic sbml
            - regulatory csv
            - cbm models objects from external packages
            - regulatory sbml
            -jsons

        The only difference between the builders and the engines is that engines read the model in five steps:
            - open
            - parse
            - read
            - close
            - clean

        Builders, on the other hand, just read and thus performing all operations at once.
        This is important for reading multiple files into a single model.

        For reading files in stages, multiple readers must be created and the director must be used to merge all the
        readers into a single model

        The only difference between the builders and the engines is that engines write the model in four steps:
                    - open
                    - write
                    - close
                    - clean

        Builders, on the other hand, just write and thus performing all operations at once.
        This is important for writing multiple files out of a single model.

        For writing files in stages, multiple writers must be created and the director must be used to merge all the
        writers into a single model

        :param io: file path, file handler or cbm model object from cobrapy or reframed
        :param config: dictionary of multiple configurations to be used when reading
        :param model: in case of writing, a GERM model must be provided
        """
        self._io = io
        self._config = config
        self._model = model
        self._dto = None
        self._variables = defaultdict(set)
        self._warnings = []

    @property
    def io(self) -> Union[str, 'TextIOWrapper', 'CobraModel', 'ReframedModel']:
        return self._io

    @property
    def config(self) -> dict:
        return self._config

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        return self._model

    @property
    @abstractmethod
    def model_type(self) -> Optional[str]:
        return

    @property
    def dto(self) -> 'DataTransferObject':
        return self._dto

    @property
    def variables(self) -> Dict[str, set]:
        return self._variables

    @property
    def warnings(self) -> list:
        return self._warnings

    @abstractmethod
    def open(self, mode: str = 'r'):
        """
        Open makes the preparation for reading or writing

        :param mode: One of the following: r - read; w - write;
        :return:
        """
        pass

    @abstractmethod
    def parse(self):
        """
        Parses a given file to a DataTransferObject

        :return:
        """
        pass

    @abstractmethod
    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables: dict = None) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        """
        Reads a model into a GERM model. If a model is provided, the read method will increment further variables
        If a variables dictionary is provided, multi-type variables can be built together with the ones available in
        the model

        Reading is performed from the middle DataTransferObject

        :param model: A valid GERM model
        :param variables: A dictionary of variables already built to be updated during reading
        :return: GERM model
        """
        pass

    @abstractmethod
    def write(self):
        """
        Writes a GERM model.
        :return:
        """

        pass

    @abstractmethod
    def close(self):
        """
        Closes the engine.
        :return:
        """
        pass

    @abstractmethod
    def clean(self):
        """
        Cleans the engine.
        :return:
        """
        pass



from itertools import product
from typing import Dict, Union, TYPE_CHECKING, Callable, Any, Type

import pandas as pd

from .algebra_utils import solution_decode, _walk
from .parsing import tokenize
from .symbolic import NoneAtom, Symbolic, Symbol

if TYPE_CHECKING:
    from mewpy.germ.variables import Gene, Metabolite, Reaction, Regulator, Interaction, Target, Variable


class Expression:

    def __init__(self,
                 symbolic: Symbolic = None,
                 variables: Dict[str, Union['Gene',
                                            'Interaction',
                                            'Metabolite',
                                            'Reaction',
                                            'Regulator',
                                            'Target',
                                            'Variable']] = None):

        """
        Expression allows high-level programming of symbolic algebra (logic and math)
        using GERM model variables (e.g. Gene, Interaction, Metabolite, Reaction, Regulator, Target, etc).

        An expression allows symbolic algebra evaluation with GERM model variables or Symbolic-based symbols using
        regular python binary operators or custom functions.

        An expression can be tokenized, translated or iterated.
        The truth table can be computed for the variables or symbols.

        The identifiers of all variables must match the names of all symbols, otherwise an error is raised.

        :param symbolic: The symbolic object which implements an algebra expression
        containing Symbolic operators, symbols or atoms. If None (default), an empty symbolic expression,
        aka NoneAtom, is created.
        :param variables: A dictionary of GERM model variables instances. The GERM model variables that must be associated
        with this symbolic expression. Variables identifiers, and thus dictionary keys,
        must match all symbols in the Symbolic expression
        """

        if symbolic is None:
            symbolic = NoneAtom()

        if variables is None:
            variables = {}

        self._symbolic: Symbolic = symbolic
        self._variables: Dict[str, Union['Gene',
                                         'Interaction',
                                         'Metabolite',
                                         'Reaction',
                                         'Regulator',
                                         'Target']] = variables

        self._link_variables_to_symbols()

    def _link_variables_to_symbols(self):

        for symbol in self.symbols.values():

            if symbol.name not in self._variables:
                raise ValueError('Expression set with incorrect variables or symbolic expression')

            else:
                symbol.model_variable = self._variables[symbol.name]

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @property
    def symbolic(self) -> Symbolic:
        """
        Symbolic property
        :return: The symbolic logic or math algebra expression.
        """
        return self._symbolic

    @property
    def variables(self) -> Dict[str, Union['Gene', 'Interaction', 'Metabolite', 'Reaction', 'Regulator', 'Target']]:
        """
        Variables dictionary property
        :return: A copy of the GERM model variables dictionary
        """
        return self._variables.copy()

    # -----------------------------------------------------------------------------
    # Static setters
    # -----------------------------------------------------------------------------
    @variables.setter
    def variables(self, value: Dict[str, Union['Gene',
                                               'Interaction',
                                               'Metabolite',
                                               'Reaction',
                                               'Regulator',
                                               'Target']]):

        """
        Variables setter
        :param value: A dictionary of GERM model variables. The GERM model variables that must be associated
        with this symbolic expression. Variables identifiers, and thus dictionary keys,
        must match all symbols in the Symbolic expression
        """

        self._variables = value

        self._link_variables_to_symbols()

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def symbols(self) -> Dict[str, 'Symbol']:

        """
        Symbols property.
        :return: A dictionary containing all symbols in the symbolic algebra expression
        """
        return {symbol.name: symbol for symbol in self.symbolic.atoms(symbols_only=True)}

    @property
    def is_none(self):
        """
        Is none property.
        :return: bool True whether the current expression is empty or a NoneAtom
        """
        return self.symbolic.is_none

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __repr__(self):
        return repr(self.symbolic)

    def __str__(self):
        return str(self.symbolic)

    def _repr_html_(self):
        """
        It returns a html representation of the gene.
        """
        return str(self.symbolic)

    def to_string(self):
        """
        String representation of the symbolic algebra expression
        :return: string with bitwise operators and symbols names
        """

        return self.symbolic.to_string()

    def to_tokenize(self):
        """
        Tokenize symbolic algebra expression
        :return: list of all tokens (no spaces) in the string representation
        """
        return tokenize(self.to_string())

    # -----------------------------------------------------------------------------
    # Iteration
    # -----------------------------------------------------------------------------
    def __iter__(self):

        return self.symbolic.__iter__()

    def __next__(self):

        return self.symbolic.__next__()

    def walk(self, reverse=False):

        """
        Iterate/walk over Symbolic algebra expression yielding Symbolic-type symbols or operators.
        Parent operators are iterated first by default.

        :param reverse: If true, Symbolic symbols are yield first
        :return: A generator object to be iterated (yield Symbolic symbols and operators)
        """

        return _walk(self.symbolic, reverse)

    def __call__(self,
                 values: Dict[str, float],
                 coefficient: float = None,
                 operators: Union[Dict[Type[Symbolic], Callable], Dict[Type[Symbolic], Any]] = None,
                 missing_value: float = 0.0,
                 decoder: dict = None,
                 **kwargs) -> Any:

        """
        Evaluate a Symbolic algebra expression based on
        the coefficients/values of the Symbolic symbols - GERM model variables.

        An expression is a also callable. Thus, the symbolic expression is evaluated according to the Symbolic
        operators and values attributed to the symbols in the values dictionary.

        :param values: A dictionary of values that the variables identifiers, aka symbols names,
        must take during expression evaluation
        :param coefficient: The value to be returned in case the expression is evaluated to True. Otherwise,
        binary values 0 or 1 are returned.
        :param operators: A dictionary of custom operators. That is, python operators-based evaluation
        (e.g. 3 > 2 yield True) can be replaced by custom callable objects such as functions. For instance,
        3 > 2 can be evaluated with max, and thus max(3, 2) yields 3 now.
        :param missing_value: If a given variable/symbol is missing in the values' dictionary, the missing
        value is used.
        :param decoder: A custom dictionary for decoding the solution (key) into a given output (value).
        Binary output is currently set to 0 or 1 by default
        :param kwargs: Additional keyword arguments for Symbolic evaluate method

        :return: The solution of the Symbolic expression evaluation as int, float or Any type.
        """
        return self.evaluate(values=values,
                             coefficient=coefficient,
                             operators=operators,
                             missing_value=missing_value,
                             decoder=decoder,
                             **kwargs)

    def evaluate(self,
                 values: Dict[str, float],
                 coefficient: float = None,
                 operators: Union[Dict[Type[Symbolic], Callable], Dict[Type[Symbolic], Any]] = None,
                 missing_value: float = 0.0,
                 decoder: dict = None,
                 **kwargs) -> Any:

        """
        Evaluate a Symbolic algebra expression based on
        the coefficients/values of the Symbolic symbols - GERM model variables.

        The symbolic expression is evaluated according to the Symbolic operators and values attributed to the symbols
        in the values' dictionary.

        :param values: A dictionary of values that the variables identifiers, aka symbols names,
        must take during expression evaluation
        :param coefficient: The value to be returned in case the expression is evaluated to True. Otherwise,
        binary values 0 or 1 are returned.
        :param operators: A dictionary of custom operators. That is, python operators-based evaluation
        (e.g. 3 > 2 yield True) can be replaced by custom callable objects such as functions. For instance,
        3 > 2 can be evaluated with max, and thus max(3, 2) yields 3 now.
        :param missing_value: If a given variable/symbol is missing in the values' dictionary, the missing
        value is used.
        :param decoder: A custom dictionary for decoding the solution (key) into a given output (value).
        Binary output is currently set to 0 or 1 by default
        :param kwargs: Additional keyword arguments for Symbolic evaluate method

        :return: The solution of the Symbolic expression evaluation as int, float or Any type.
        """
        res = self.symbolic.evaluate(values=values, operators=operators, default=missing_value, **kwargs)

        res = solution_decode(res, decoder)

        # if a coefficient is provided and the expression is evaluated to true, the respective coefficient will
        # be returned
        if coefficient is None:
            return res

        else:

            if res:
                return coefficient

            return res

    def truth_table(self,
                    values: Dict[str, float] = None,
                    strategy: str = 'max',
                    coefficient: float = None,
                    operators: Union[Dict[Type[Symbolic], Callable], Dict[Type[Symbolic], Any]] = None,
                    decoder: Dict[Any, Any] = None) -> pd.DataFrame:
        """
        It calculates the truth table for this expression. The truth table is composed by the combination of values
        taken by empty, numeric and symbolic variables available in the algebra expression.

        The symbolic expression is evaluated according to the Symbolic operators and values attributed to the symbols
        in the values' dictionary.

        :param values: A dictionary of values that the variables identifiers, aka symbols names,
        can take during expression evaluation. If the values dictionary is not provided,
        variables' values/states are retrieved from the variables' coefficients using the strategy
        defined by the strategy's parameter.
        :param strategy: The truth table can be calculated using the maximum or minimum value
        in the variables' coefficients. Otherwise, the truth table is calculated using all variables' coefficients.
        :param coefficient: The value to be returned in case the expression is evaluated to True. Otherwise,
        binary values 0 or 1 are returned.
        :param operators: A dictionary of custom operators. That is, python operators-based evaluation
        (e.g. 3 > 2 yield True) can be replaced by custom callable objects such as functions. For instance,
        3 > 2 can be evaluated with max, and thus max(3, 2) yields 3 now.
        :param decoder: A custom dictionary for decoding the solution (key) into a given output (value).
        Binary output is currently set to 0 or 1 by default

        :return: The solution of the combination of values taken by empty, numeric and symbolic variables available
        in the algebra expression as a pandas DataFrame. DataFrame columns should be composed by all
        empty, numeric and symbolic variables in addition to the result columns
        (the result of evaluating the algebra expression). DataFrame rows should stand for the cartesian product of all
        combination of values taken by empty, numeric and symbolic variables,
        unless specific parameters are taken.
        """
        if not values:
            values = {}

        state = {}
        for var_id, var in self.variables.items():
            if var_id in values:
                state[var_id] = values[var_id]

            else:
                if var.is_metabolite() and var.exchange_reaction is not None and var.exchange_reaction.id in values:
                    state[var_id] = values[var.exchange_reaction.id]
                else:
                    if strategy == 'max':
                        state[var_id] = max(var.coefficients)
                    elif strategy == 'min':
                        state[var_id] = min(var.coefficients)
                    else:
                        state[var_id] = var.coefficients

        truth_table = []
        if strategy in ('max', 'min'):
            state['result'] = self.evaluate(values=state,
                                            coefficient=coefficient,
                                            operators=operators,
                                            decoder=decoder)
            truth_table.append(state)

        elif strategy == 'all':
            variables = list(self.variables.keys())

            for mid_state in product(*state.values()):
                mid_state = dict(list(zip(variables, mid_state)))

                mid_state['result'] = self.evaluate(values=mid_state,
                                                    coefficient=coefficient,
                                                    operators=operators,
                                                    decoder=decoder)
                truth_table.append(mid_state)

        else:
            raise ValueError('coefficients must be max, min or all')

        return pd.DataFrame(truth_table)



from abc import ABC, abstractmethod
import signal
import sys
from ..util.constants import EAConstants
from ..util.process import cpu_count


class SolutionInterface(ABC):
    """ An interface for EA solutions.
    """

    @abstractmethod
    def get_fitness(self):
        """
        returns a list of fitness values
        """
        raise NotImplementedError

    def get_representation(self):
        """
        returns a set representation of the solution
        """
        raise NotImplementedError


class Solution(SolutionInterface):

    def __init__(self, values, fitness, constraints=None, is_maximize=True):
        """
        EA Solution

        :param values: Representation of the solution.
        :param fitness:  A list of fitness values.
        :param constraints: Decoding of the representation into metabolic constraints.
        :param is_maximize: If the solution results from a maximization problem.

        """
        self.values = values
        self.fitness = fitness
        self.constraints = {} if constraints is None else constraints
        self._is_maximize = is_maximize

    def get_fitness(self):
        return self.fitness

    def get_representation(self):
        return self.values

    def get_constraints(self):
        return self.constraints

    def __str__(self):
        return f"{self.fitness};{self.values}"

    def __repr__(self):
        return f"{self.fitness};{self.values}"

    def __eq__(self, solution):
        return set(self.values) == set(solution.values)

    def __gt__(self, solution):
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=self._is_maximize) == 1
        return False

    def __lt__(self, solution):
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=self._is_maximize) == -1
        return False

    def __ge__(self, solution):
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=self._is_maximize) != -1
        return False

    def __le__(self, solution):
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=self._is_maximize) != 1
        return False

    def __copy__(self):
        import copy
        values = copy.copy(self.values)
        fitness = self.fitness.copy()
        new_solution = Solution(values, fitness)
        return new_solution

    def __hash__(self):
        return hash(str(set(self.values)))

    def toDict(self):
        d = {'values': self.values,
             'fitness': self.fitness,
             'constraints': self.constraints}
        return d


class AbstractEA(ABC):

    def __init__(self, problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS,
                 mp=True, visualizer=False, **kwargs):

        self.problem = problem
        self.initial_population = initial_population
        self.max_generations = max_generations
        self.visualizer = visualizer
        self.mp = mp
        self.final_population = None

    def run(self, simplify=True):
        """ Runs the optimization for the defined problem.
        The number of objectives is defined to be the number of evaluation functions in fevalution.
        """
        # Register signal handler for linux
        signal.signal(signal.SIGINT, self.__signalHandler)

        if self.problem.fevaluation is None or len(self.problem.fevaluation) == 0:
            raise ValueError("At leat one objective should be provided.")

        # builds the target list
        self.problem.pre_process()

        if self.problem.number_of_objectives == 1:
            final_pop = self._run_so()
        else:
            final_pop = self._run_mo()
        pop = self._convertPopulation(final_pop)
        pop = filter_duplicates(pop)
        if simplify:
            n_cpu = cpu_count() if self.mp else 1
            pop = self.problem.simplify_population(pop, n_cpu)
        self.final_population = pop
        return self.final_population

    def dataframe(self):
        """Returns a dataframe of the final population.

        :raises Exception: if the final population is empty or None.
        :return: Returns a dataframe of the final population
        :rtype: pandas.Dataframe
        """
        if not self.final_population:
            raise Exception("No solutions")
        table = [[x.values, len(x.values)]+x.fitness for x in self.final_population]
        import pandas as pd
        columns = ["Modification", "Size"]
        columns.extend([obj.short_str() for obj in self.problem.fevaluation])
        df = pd.DataFrame(table, columns=columns)
        return df

    def plot(self):
        """Plots the final population.

        :raises Exception:  if the final population is empty or None.
        """
        if not self.final_population:
            raise Exception("No solutions")
        from ..visualization.plot import StreamingPlot
        labels = [obj.short_str() for obj in self.problem.fevaluation]
        p = StreamingPlot(axis_labels=labels)
        p.plot(self.final_population)

    def __signalHandler(self, signum, frame):
        if EAConstants.KILL_DUMP:
            print("Dumping current population.")
            try:
                pop = self._get_current_population()
                data = [s.toDict() for s in pop]
                import json
                from datetime import datetime
                now = datetime.now()
                dt_string = now.strftime("%d%m%Y-%H%M%S")
                with open(f'mewpy-dump-{dt_string}.json', 'w') as outfile:
                    json.dump(data, outfile)
            except Exception:
                print("Unable to dump population.")
            print("Exiting")
        sys.exit(0)

    @ abstractmethod
    def _convertPopulation(self, population):
        raise NotImplementedError

    @ abstractmethod
    def _run_so(self):
        raise NotImplementedError

    @ abstractmethod
    def _run_mo(self):
        raise NotImplementedError

    @ abstractmethod
    def _get_current_population(self):
        raise NotImplementedError


def dominance_test(solution1, solution2, maximize=True):
    """
    Testes Pareto dominance

    :param solution1: The first solution.
    :param solution2: The second solution.
    :param maximize: (bool) maximization (True) or minimization (False)
    :returns: 1 : if the first solution dominates the second; -1 : if the second solution dominates the first; \
         0 : if non of the solutions dominates the other.

    """

    best_is_one = 0
    best_is_two = 0

    if isinstance(solution1, list):
        s1 = solution1
    else:
        s1 = solution1.fitness

    if isinstance(solution2, list):
        s2 = solution2
    else:
        s2 = solution2.fitness

    for i in range(len(s1)):
        value1 = s1[i]
        value2 = s2[i]
        if value1 != value2:
            if value1 < value2:
                best_is_one = 1
            if value1 > value2:
                best_is_two = 1

    if best_is_one > best_is_two:
        result = 1
    elif best_is_two > best_is_one:
        result = -1
    else:
        result = 0

    if not maximize:
        result = -1 * result

    return result


def non_dominated_population(solutions, maximize=True, filter_duplicate=True):
    """
    Returns the non dominated solutions from the population.
    """
    # number of solutions dominating solution ith
    dominating_ith = [0 for _ in range(len(solutions))]

    # list of solutions dominated by solution ith
    ith_dominated = [[] for _ in range(len(solutions))]

    front = []

    for p in range(len(solutions) - 1):
        for q in range(p + 1, len(solutions)):
            dominance_test_result = dominance_test(solutions[p], solutions[q], maximize=maximize)

            if dominance_test_result == -1:
                ith_dominated[p].append(q)
                dominating_ith[q] += 1
            elif dominance_test_result == 1:
                ith_dominated[q].append(p)
                dominating_ith[p] += 1

    for i in range(len(solutions)):
        if dominating_ith[i] == 0:
            front.append(solutions[i])

    if filter_duplicate:
        result = filter_duplicates(front)
    else:
        result = front
    return result


def filter_duplicates(population):
    """ Filters equal solutions from a population
    """

    def remove_equal(individual, population):
        filtered = []
        for other in population:
            if individual != other:
                filtered.append(other)
        return filtered

    fitered_list = []
    p = population
    while len(p) > 1:
        individual = p[0]
        fitered_list.append(individual)
        p = remove_equal(individual, p)
    if p:
        fitered_list.extend(p)
    return fitered_list


def cmetric(pf1, pf2, maximize=True):
    """
    Computes the c-metric quality indicator.

    :param pf1: The first pareto front.
    :param pf2: The second pareto front.
    :param maximize: (bool) maximization (True) or minimization (False).
    :returns: r1,r2,pf1_2,pf2_1
                r1: percentage of solutions on pf2 dominated by some solution on pf1;
                r2: percentage of solutions on pf1 dominated by some solution on pf2;
                pf1_2: solutions on pf2 dominated by some solution on pf1;
                pf2_1: solutions on pf1 dominated by some solution on pf2.

    """
    # solutions on pf2 dominated by some solution on pf1
    pf1_2 = set()
    # solutions on pf1 dominated by some solution on pf2
    pf2_1 = set()
    for s1 in pf1:
        for s2 in pf2:
            d = dominance_test(s1, s2, maximize=maximize)
            if d == 1:
                pf1_2.add(s2)
            elif d == -1:
                pf2_1.add(s1)
    r1 = len(pf1_2) / len(pf2)
    r2 = len(pf2_1) / len(pf1)
    return r1, r2, pf1_2, pf2_1
